/*
 * antibioticsimulationgrid.cpp
 *
 * Created by epienaar on 06/15/2015.
 */


#include "antibioticsimulationgrid.h"
#include "stat.h"
#include "grdiffusionFFT.h"
#include "antibioticconcentration.h"
#include "grgrid.h"
#include <string>
#include <sstream>

using namespace std;

const std::string AntibioticSimulationGrid::antibioticStatHeaders[] = {// Global antibiotic stats
                                                                       "_Blood",
                                                                       "_NormalTissue",
                                                                       "_Granuloma",
                                                                       "_Caseum",
                                                                       "_PartialCaseum",
                                                                       "_WholeGrid",
                                                                       "_Lung",
                                                                       "_Periph",
                                                                       "_IntMac",

                                                                       // AUC antibiotic stats.
                                                                       "_AUC_Blood",
                                                                       "_AUC_NormalTissue",
                                                                       "_AUC_Granuloma",
                                                                       "_AUC_Caseum",
                                                                       "_AUC_WholeGrid",
                                                                       "_AUC_Lung",
                                                                       "_AUC_Periph",

                                                                       // Cmax antibiotic stats.
                                                                       "_Cmax_Blood",
                                                                       "_Cmax_NormalTissue",
                                                                       "_Cmax_Granuloma",
                                                                       "_Cmax_Caseum",
                                                                       "_Cmax_WholeGrid",

                                                                       // Antibiotic distribution stats
                                                                       "_min",
                                                                       "_max",

                                                                       // Single drug kill stats
                                                                       "_IntMtbSingleDrugKill",
                                                                       "_ExtMtbSingleDrugKill",
                                                                       "_NonrepSingleDrugKill",

                                                                       // Antibiotic MIC stats.
                                                                       "_Comp>MIC",
                                                                       "_ExtMtb>MIC",
                                                                       "_Macs>MIC",
                                                                       "_IntMtb>MIC",

                                                                       // Antibiotic Cell Stats
                                                                       "_AvgInfInt",
                                                                       "_AvgExtMtb",
                                                                       "_AvgNonMtb",

                                                                       // Antibiotic resistance risk Stats
                                                                       "_AtRiskIntMtb",
                                                                       "_AtRiskExtMtb",
                                                                       "_AtRiskNonrep",
                                                                       "_AtRiskMacs",
                                                                       "_AtRiskComps",

                                                                       // Antibiotic resistance Stats
                                                                       "_ResistantTotExtMtb",
                                                                       "_ResistantTotNonRepExtMtb",
                                                                       "_ResistantTotIntMtb",
                                                                       "_ResistantTotMtb",

                                                                       "_MdrTotExtMtb",
                                                                       "_MdrTotNonRepExtMtb",
                                                                       "_MdrTotIntMtb",
                                                                       "_MdrTotMtb",

                                                                       // Antibiotic resistance compensatory mutation stats
                                                                       "_compensatedResistantTotExtMtb",
                                                                       "_compensatedResistantTotNonRepExtMtb",
                                                                       "_compensatedResistantTotIntMtb",
                                                                       "_compensatedResistantTotMtb",
                                                                       "_mdrCompensatedResistantTotExtMtb",
                                                                       "_mdrCompensatedResistantTotNonRepExtMtb",
                                                                       "_mdrCompensatedResistantTotIntMtb",
                                                                       "_mdrCompensatedResistantTotMtb",

                                                                       "_mdrDoubleCompensatedResistantTotExtMtb",
                                                                       "_mdrDoubleCompensatedResistantTotNonRepExtMtb",
                                                                       "_mdrDoubleCompensatedResistantTotIntMtb",
                                                                       "_mdrDoubleCompensatedResistantTotMtb",

                                                                       // Cumulative Antibiotic resistance Stats
                                                                       "_CumResistantTotExtMtb",
                                                                       "_CumResistantTotNonRepExtMtb",
                                                                       "_CumResistantTotIntMtb",
                                                                       "_CumResistantTotMtb",

                                                                       "_CumMdrTotExtMtb",
                                                                       "_CumMdrTotNonRepExtMtb",
                                                                       "_CumMdrTotIntMtb",
                                                                       "_CumMdrTotMtb",

                                                                       // Cumulative Antibiotic resistance compensatory mutation stats
                                                                       "_CumCompensatedResistantTotExtMtb",
                                                                       "_CumCompensatedResistantTotNonRepExtMtb",
                                                                       "_CumCompensatedResistantTotIntMtb",
                                                                       "_CumCompensatedResistantTotMtb",
                                                                       "_CumMdrCompensatedResistantTotExtMtb",
                                                                       "_CumMdrCompensatedResistantTotNonRepExtMtb",
                                                                       "_CumMdrCompensatedResistantTotIntMtb",
                                                                       "_CumMdrCompensatedResistantTotMtb",

                                                                       "_CumMdrDoubleCompensatedResistantTotExtMtb",
                                                                       "_CumMdrDoubleCompensatedResistantTotNonRepExtMtb",
                                                                       "_CumMdrDoubleCompensatedResistantTotIntMtb",
                                                                       "_CumMdrDoubleCompensatedResistantTotMtb",

                                                                       // Efflux pump stats
                                                                       "_effluxInducedExtMtb",
                                                                       "_effluxInducedNonRepExtMtb",
                                                                       "_effluxInducedIntMtb",
                                                                       "_effluxInducedTotMtb",

                                                                       "_doubleEffluxInducedExtMtb",
                                                                       "_doubleEffluxInducedNonRepExtMtb",
                                                                       "_doubleEffluxInducedIntMtb",
                                                                       "_doubleEffluxInducedTotMtb"

                                                                      };

const int AntibioticSimulationGrid::nrHeaders = sizeof(AntibioticSimulationGrid::antibioticStatHeaders) / sizeof(std::string);

AntibioticGrid::AntibioticGrid()
    :
      _grid()
    , _u_grid()
    , _v_grid()
    , _casBnd_grid()
    , _epithBnd_grid()
    , _dim()
{
}

AntibioticGrid::AntibioticGrid(const Pos& dim)
    :
      _grid(dim.paddedSize())
    , _u_grid(dim.paddedSize())
    , _v_grid(dim.paddedSize())
    , _casBnd_grid(dim.size())
    , _epithBnd_grid(dim.size())
    , _dim(dim)
{
}

AntibioticGrid::AntibioticGrid(const AntibioticGrid& ag)
  : _grid(ag._grid)
  , _u_grid(ag._u_grid)
  , _v_grid(ag._v_grid)
  , _casBnd_grid(ag._casBnd_grid)
  , _epithBnd_grid(ag._epithBnd_grid)
  , _dim(ag._dim)
{

}

AntibioticGrid::~AntibioticGrid()
{
}

DrugMtbStat::DrugMtbStat()
    : _extMtb(0)
    , _nonrepExtMtb(0)
    , _intMtb(0)
    , _totMtb(0)
{
}

DrugMtbStat::~DrugMtbStat()
{
}

AntibioticSimulationGrid::AntibioticSimulationGrid()
    :
        _dim()
      , _name("")                // abbreviation of the antibiotic
      , _cutOffValue(DIFFUSION_CUTOFF)

      // Assign antibiotic specific parameters from param file - see params.def for details
      , _PKModel(0)
      , _vesselActivity(0.0)
      , _partitionUsingSteadyState(0)
      , _degRateConst(0.0)
      , _degRateConstInt(0.0)
      , _diffusivity(0.0)
      , _minDiffusivity(0.0)
      , _kAbs(0.0)
      , _q(0.0)
      , _volDistCentral(0.0)     // This is assuming a 2 compartment model as for INH and RIF - will need to be modified if other drugs have different PK structures
      , _volDistPeriph(0.0)      // This is assuming a 2 compartment model as for INH and RIF - will need to be modified if other drugs have different PK structures
      , _clearance(0.0)
      , _dose(0.0)
      , _dosageInterval(0.0)
      , _cellUptake(0.0)
      , _vascularPermeability(0.0)
      , _permCoeff(0.0)
      , _intraActivity(0.0)
      , _extraActivity(0.0)
      , _nonrepActivity(0.0)
      , _c50Extra(0.0)
      , _c50Intra(0.0)
      , _c50Nonrep(0.0)
      , _consumptionRate(0.0)
      , _consumptionHalfSat(0.0)
      , _mic(0.0)
      , _hillExtra(0.0)
      , _hillIntra(0.0)
      , _hillNonrep(0.0)
      , _caseumUnboundFraction(0.0)
      , _caseumBindingRateConst(0.0)
      , _epitheliumBindingKa(0.0)
      , _epitheliumBindingRateConst(0.0)
      , _cellularKout(0.0)
      , _MW(0.0)
      , _resistRiskIndicatorFlag(0)
      , _aucRiskLB(0.0)
      , _aucRiskUB(0.0)
      , _cmaxRiskLB(0.0)
      , _cmaxRiskUB(0.0)
      , _numberResistMutations(0.0)
      , _relativeFitnessResist(0.0)
      , _numberCompensatoryMutations(0.0)
      , _foldIncreaseC50FromResistance(0.0)
      , _numTimeStepsEffluxInduction(0)
      , _minEffluxInduction(0.0)
      , _maxEffluxInduction(0.0)
      , _concentrationChangeByEfflux(0.0)
      , _numberOfRandomMonoResistantBugsToAdd(0.0)
      , _volDistLung(0.0)

      // Concentration Grids
      , _pMat()

      #if DIMENSION == 2
      // Crank-Nicolson ADI is a 2D only diffusion algorithm.
      , _cnadiChemical(NULL)
      #endif

      , _currentGrid(new AntibioticGrid())
      , _nextGrid(new AntibioticGrid())

      // Vascular variables
      , _bloodConcentration(0.0)
      , _absComp1Concentration(0.0)
      , _absComp2Concentration(0.0)
      , _perCompConcentration(0.0)
      , _lungCompConcentration(0.0)
      , _intMacConcentration(0.0)

      , _initvector(0.0, 5)  // valarrays for numerical methods -- length set to highest for drug delivery option -- will not be used for std drug dynamics

      // Stats Grids
      , _intracellularGrid()
      , _totalGrid()
      , _aucIntraGrid()
      , _aucExtraGrid()
      , _aucTotalGrid()
      , _cmaxIntraGrid()
      , _cmaxExtraGrid()
      , _cmaxTotalGrid()

      // Stats
      , _totalWholeGrid(0.0)
      , _totalGran(0.0)
      , _totalNorm(0.0)
      , _totalCaseum(0.0)
      , _partialCaseum(0.0)
      , _avgConcInfInt(0.0)
      , _avgConcExtMtb(0.0)
      , _avgConcNonMtb(0.0)

      , _drugConcentrationArea(0.0)
      , _drugConcentrationAreaNorm(0.0)
      , _drugConcentrationAreaCaseum(0.0)
      , _drugConcentrationAreaGran(0.0)
      , _drugConcentrationAreaBlood(0.0)
      , _drugConcentrationAreaLung(0.0)
      , _drugConcentrationAreaPeriph(0.0)
      , _cmax(0.0)
      , _cmaxNorm(0.0)
      , _cmaxCaseum(0.0)
      , _cmaxGran(0.0)
      , _cmaxBlood(0.0)
      , _compOverMIC(0.0)
      , _extMtbOverMIC(0.0)
      , _macsOverMIC(0.0)
      , _intMtbOverMIC(0.0)
      , _intMtbHighRiskOfResistance(0.0)
      , _extMtbHighRiskOfResistance(0.0)
      , _nonrepMtbHighRiskOfResistance(0.0)
      , _macsHighRiskOfResistance(0.0)
      , _compHighRiskOfResistance(0.0)
      , _intMtbSingleDrugKill(0.0)
      , _extMtbSingleDrugKill(0.0)
      , _nonrepSingleDrugKill(0.0)

      , _resistant()
      , _mdr()
      , _compensatedResistant()
      , _mdrCompensatedResistant()
      , _mdrDoubleCompensatedResistant()
      , _cumResistant()
      , _cumMdr()
      , _cumCompensated()
      , _cumMdrCompensatedResistant()
      , _cumMdrDoubleCompensated()
      , _effluxInduced()
      , _doubleEffluxInduced()

      , _drugConcentration()
      , _drugConcentrationNorm()
      , _drugConcentrationCaseum()
      , _drugConcentrationGran()
      , _drugConcentrationblood()
      , _drugConcentrationLung()
      , _drugConcentrationPeriph()

      , _drugDistStat()

{
}

AntibioticSimulationGrid::AntibioticSimulationGrid(const string& antibioticName,
                                                   const Pos& dim,
                                                   const boost::property_tree::ptree& pt,
                                                   GrDiffusion* pDiffusion, double dt
                                                   )
    :
      _dim(dim)
    , _name(antibioticName)                // abbreviation of the antibiotic
    , _cutOffValue(DIFFUSION_CUTOFF)

      // Assign antibiotic specific parameters from param file - see params.def for details
    , _PKModel(0)
    , _vesselActivity(0.0)
    , _partitionUsingSteadyState(0)
    , _degRateConst(0.0)
    , _degRateConstInt(0.0)
    , _diffusivity(0.0)
    , _minDiffusivity(0.0)

    , _kAbs(0.0)
    , _q(0.0)
    , _volDistCentral(0.0)     // This is assuming a 2 compartment model as for INH and RIF - will need to be modified if other drugs have different PK structures
    , _volDistPeriph(0.0)      // This is assuming a 2 compartment model as for INH and RIF - will need to be modified if other drugs have different PK structures
    , _clearance(0.0)

    , _dose(0.0)
    , _dosageInterval(0.0)
    , _cellUptake(0.0)
    , _vascularPermeability(0.0)
    , _permCoeff(0.0)
    , _intraActivity(0.0)
    , _extraActivity(0.0)
    , _nonrepActivity(0.0)
    , _c50Extra(0.0)
    , _c50Intra(0.0)
    , _c50Nonrep(0.0)
    , _consumptionRate(0.0)
    , _consumptionHalfSat(0.0)
    , _mic(0.0)
    , _hillExtra(0.0)
    , _hillIntra(0.0)
    , _hillNonrep(0.0)
    , _caseumUnboundFraction(0.0)
    , _caseumBindingRateConst(0.0)
    , _epitheliumBindingKa(0.0)
    , _epitheliumBindingRateConst(0.0)
    , _cellularKout(0.0)
    , _MW(0.0)
    , _resistRiskIndicatorFlag(0.0)
    , _aucRiskLB(0.0)
    , _aucRiskUB(0.0)
    , _cmaxRiskLB(0.0)
    , _cmaxRiskUB(0.0)
    , _numberResistMutations(0.0)
    , _relativeFitnessResist(0.0)
    , _numberCompensatoryMutations(0.0)
    , _foldIncreaseC50FromResistance(0.0)
    , _numTimeStepsEffluxInduction(0.0)
    , _minEffluxInduction(0.0)
    , _maxEffluxInduction(0.0)
    , _concentrationChangeByEfflux(0.0)
    , _numberOfRandomMonoResistantBugsToAdd(0.0)
    , _volDistLung(0.0)

    //Concentration Grids
    , _pMat()

    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    , _cnadiChemical(NULL)
    #endif

    , _currentGrid(new AntibioticGrid(dim))
    , _nextGrid(new AntibioticGrid(dim))

    // Vascular variables
    , _bloodConcentration(0.0)
    , _absComp1Concentration(0.0)
    , _absComp2Concentration(0.0)
    , _perCompConcentration(0.0)
    , _lungCompConcentration(0.0)
    , _intMacConcentration(0.0)

    , _initvector(0.0, 5)  // valarrays for numerical methods -- length set to highest for drug delivery option -- will not be used for std drug dynamics

    // Stats Grids
    , _intracellularGrid(dim.size())
    , _totalGrid(dim.size())
    , _aucIntraGrid(dim.size())
    , _aucExtraGrid(dim.size())
    , _aucTotalGrid(dim.size())
    , _cmaxIntraGrid(dim.size())
    , _cmaxExtraGrid(dim.size())
    , _cmaxTotalGrid(dim.size())

    // Stats
    , _totalWholeGrid(0.0)
    , _totalGran(0.0)
    , _totalNorm(0.0)
    , _totalCaseum(0.0)
    , _partialCaseum(0.0)
    , _avgConcInfInt(0.0)
    , _avgConcExtMtb(0.0)
    , _avgConcNonMtb(0.0)

    , _drugConcentrationArea(0.0)
    , _drugConcentrationAreaNorm(0.0)
    , _drugConcentrationAreaCaseum(0.0)
    , _drugConcentrationAreaGran(0.0)
    , _drugConcentrationAreaBlood(0.0)
    , _drugConcentrationAreaLung(0.0)
    , _drugConcentrationAreaPeriph(0.0)
    , _cmax(0.0)
    , _cmaxNorm(0.0)
    , _cmaxCaseum(0.0)
    , _cmaxGran(0.0)
    , _cmaxBlood(0.0)
    , _compOverMIC(0.0)
    , _extMtbOverMIC(0.0)
    , _macsOverMIC(0.0)
    , _intMtbOverMIC(0.0)
    , _intMtbHighRiskOfResistance(0.0)
    , _extMtbHighRiskOfResistance(0.0)
    , _nonrepMtbHighRiskOfResistance(0.0)
    , _macsHighRiskOfResistance(0.0)
    , _compHighRiskOfResistance(0.0)
    , _intMtbSingleDrugKill(0.0)
    , _extMtbSingleDrugKill(0.0)
    , _nonrepSingleDrugKill(0.0)

    , _resistant()
    , _mdr()
    , _compensatedResistant()
    , _mdrCompensatedResistant()
    , _mdrDoubleCompensatedResistant()
    , _cumResistant()
    , _cumMdr()
    , _cumCompensated()
    , _cumMdrCompensatedResistant()
    , _cumMdrDoubleCompensated()
    , _effluxInduced()
    , _doubleEffluxInduced()

    , _drugConcentration(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)
    , _drugConcentrationNorm(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)
    , _drugConcentrationCaseum(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)
    , _drugConcentrationGran(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)
    , _drugConcentrationblood(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)
    , _drugConcentrationLung(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)
    , _drugConcentrationPeriph(_PARAM(Drugs_drugConcentrationStatInterval), 0.0)

    , _drugDistStat()

{
    loadParameters(pt, dt);
    setupDiffusion(pDiffusion, dim);
    initFromParameterFile(pt);
}

AntibioticSimulationGrid::AntibioticSimulationGrid(const AntibioticSimulationGrid& asg)
    : _dim(asg._dim)
    , _name(asg._name)
    , _cutOffValue(asg._cutOffValue)
    , _PKModel(asg._PKModel)
    , _vesselActivity(asg._vesselActivity)
    , _partitionUsingSteadyState(asg._partitionUsingSteadyState)
    , _degRateConst(asg._degRateConst)
    , _degRateConstInt(asg._degRateConstInt)
    , _diffusivity(asg._diffusivity)
    , _minDiffusivity(asg._minDiffusivity)

    , _kAbs(asg._kAbs)
    , _q(asg._q)
    , _volDistCentral(asg._volDistCentral)
    , _volDistPeriph(asg._volDistPeriph)
    , _clearance(asg._clearance)

    , _dose(asg._dose)
    , _dosageInterval(asg._dosageInterval)
    , _cellUptake(asg._cellUptake)
    , _vascularPermeability(asg._vascularPermeability)
    , _permCoeff(asg._permCoeff)
    , _intraActivity(asg._intraActivity)
    , _extraActivity(asg._extraActivity)
    , _nonrepActivity(asg._nonrepActivity)
    , _c50Extra(asg._c50Extra)
    , _c50Intra(asg._c50Intra)
    , _c50Nonrep(asg._c50Nonrep)
    , _consumptionRate(asg._consumptionRate)
    , _consumptionHalfSat(asg._consumptionHalfSat)
    , _mic(asg._mic)
    , _hillExtra(asg._hillExtra)
    , _hillIntra(asg._hillIntra)
    , _hillNonrep(asg._hillNonrep)
    , _caseumUnboundFraction(asg._caseumUnboundFraction)
    , _caseumBindingRateConst(asg._caseumBindingRateConst)
    , _epitheliumBindingKa(asg._epitheliumBindingKa)
    , _epitheliumBindingRateConst(asg._epitheliumBindingRateConst)
    , _cellularKout(asg._cellularKout)
    , _MW(asg._MW)
    , _resistRiskIndicatorFlag(asg._resistRiskIndicatorFlag)
    , _aucRiskLB(asg._aucRiskLB)
    , _aucRiskUB(asg._aucRiskUB)
    , _cmaxRiskLB(asg._cmaxRiskLB)
    , _cmaxRiskUB(asg._cmaxRiskUB)
    , _numberResistMutations(asg._numberResistMutations)
    , _relativeFitnessResist(asg._relativeFitnessResist)
    , _numberCompensatoryMutations(asg._numberCompensatoryMutations)
    , _foldIncreaseC50FromResistance(asg._foldIncreaseC50FromResistance)
    , _numTimeStepsEffluxInduction(asg._numTimeStepsEffluxInduction)
    , _minEffluxInduction(asg._minEffluxInduction)
    , _maxEffluxInduction(asg._maxEffluxInduction)
    , _concentrationChangeByEfflux(asg._concentrationChangeByEfflux)
    , _numberOfRandomMonoResistantBugsToAdd(asg._numberOfRandomMonoResistantBugsToAdd)
    , _volDistLung(asg._volDistLung)

    //Concentration Grids
    , _pMat(asg._pMat)

    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    , _cnadiChemical(asg._cnadiChemical ? new CrankNicolsonADIChemical(*(asg._cnadiChemical)) : NULL)
    #endif

    , _currentGrid(new AntibioticGrid(*(asg._currentGrid)))
    , _nextGrid(new AntibioticGrid(*(asg._nextGrid)))

    // Vascular variables
    , _bloodConcentration(asg._bloodConcentration)
    , _absComp1Concentration(asg._absComp1Concentration)
    , _absComp2Concentration(asg._absComp2Concentration)
    , _perCompConcentration(asg._perCompConcentration)
    , _lungCompConcentration(asg._lungCompConcentration)
    , _intMacConcentration(asg._intMacConcentration)

    , _initvector(asg._initvector)

    // Stats Grids
    , _intracellularGrid(asg._intracellularGrid)
    , _totalGrid(asg._totalGrid)
    , _aucIntraGrid(asg._aucIntraGrid)
    , _aucExtraGrid(asg._aucExtraGrid)
    , _aucTotalGrid(asg._aucTotalGrid)
    , _cmaxIntraGrid(asg._cmaxIntraGrid)
    , _cmaxExtraGrid(asg._cmaxExtraGrid)
    , _cmaxTotalGrid(asg._cmaxTotalGrid)

    // Stats
    , _totalWholeGrid(asg._totalWholeGrid)
    , _totalGran(asg._totalGran)
    , _totalNorm(asg._totalNorm)
    , _totalCaseum(asg._totalCaseum)
    , _partialCaseum(asg._partialCaseum)
    , _avgConcInfInt(asg._avgConcInfInt)
    , _avgConcExtMtb(asg._avgConcExtMtb)
    , _avgConcNonMtb(asg._avgConcNonMtb)

    , _drugConcentrationArea(asg._drugConcentrationArea)
    , _drugConcentrationAreaNorm(asg._drugConcentrationAreaNorm)
    , _drugConcentrationAreaCaseum(asg._drugConcentrationAreaCaseum)
    , _drugConcentrationAreaGran(asg._drugConcentrationAreaGran)
    , _drugConcentrationAreaBlood(asg._drugConcentrationAreaBlood)
    , _drugConcentrationAreaLung(asg._drugConcentrationAreaLung)
    , _drugConcentrationAreaPeriph(asg._drugConcentrationAreaPeriph)
    , _cmax(asg._cmax)
    , _cmaxNorm(asg._cmaxNorm)
    , _cmaxCaseum(asg._cmaxCaseum)
    , _cmaxGran(asg._cmaxGran)
    , _cmaxBlood(asg._cmaxBlood)
    , _compOverMIC(asg._compOverMIC)
    , _extMtbOverMIC(asg._extMtbOverMIC)
    , _macsOverMIC(asg._macsOverMIC)
    , _intMtbOverMIC(asg._intMtbOverMIC)
    , _intMtbHighRiskOfResistance(asg._intMtbHighRiskOfResistance)
    , _extMtbHighRiskOfResistance(asg._extMtbHighRiskOfResistance)
    , _nonrepMtbHighRiskOfResistance(asg._nonrepMtbHighRiskOfResistance)
    , _macsHighRiskOfResistance(asg._macsHighRiskOfResistance)
    , _compHighRiskOfResistance(asg._compHighRiskOfResistance)
    , _intMtbSingleDrugKill(asg._intMtbSingleDrugKill)
    , _extMtbSingleDrugKill(asg._extMtbSingleDrugKill)
    , _nonrepSingleDrugKill(asg._nonrepSingleDrugKill)

    , _resistant(asg._resistant)
    , _mdr(asg._mdr)
    , _compensatedResistant(asg._compensatedResistant)
    , _mdrCompensatedResistant(asg._mdrCompensatedResistant)
    , _mdrDoubleCompensatedResistant(asg._mdrDoubleCompensatedResistant)
    , _cumResistant(asg._cumResistant)
    , _cumMdr(asg._cumMdr)
    , _cumCompensated(asg._cumCompensated)
    , _cumMdrCompensatedResistant(asg._cumMdrCompensatedResistant)
    , _cumMdrDoubleCompensated(asg._cumMdrDoubleCompensated)
    , _effluxInduced(asg._effluxInduced)
    , _doubleEffluxInduced(asg._doubleEffluxInduced)

    , _drugConcentration(asg._drugConcentration)
    , _drugConcentrationNorm(asg._drugConcentrationNorm)
    , _drugConcentrationCaseum(asg._drugConcentrationCaseum)
    , _drugConcentrationGran(asg._drugConcentrationGran)
    , _drugConcentrationblood(asg._drugConcentrationblood)
    , _drugConcentrationLung(asg._drugConcentrationLung)
    , _drugConcentrationPeriph(asg._drugConcentrationPeriph)

    , _drugDistStat(asg._drugDistStat)
{
}

AntibioticSimulationGrid::~AntibioticSimulationGrid()
{
    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    delete _cnadiChemical;
    #endif

    delete _currentGrid;
    delete _nextGrid;
}

void AntibioticSimulationGrid::loadParameters(const boost::property_tree::ptree& pt, double dt)
{

    string drugNodePath = std::string(PARAM_FILE_ROOT_NODE_NAME) + ".Drugs." + _name;
    boost::optional<const boost::property_tree::ptree&> drugptOptional = pt.get_child_optional(drugNodePath);

    if(drugptOptional)
    {
        const boost::property_tree::ptree& drugParams = (*drugptOptional).get_child("<xmlattr>");

        bool result = true;
        result &= getDrugParam(drugParams, drugNodePath, "PKModel", _PKModel);
        result &= getDrugParam(drugParams, drugNodePath, "vesselActivity", _vesselActivity);
        result &= getDrugParam(drugParams, drugNodePath, "partitionUsingSteadyState", _partitionUsingSteadyState);
        result &= getDrugParam(drugParams, drugNodePath, "degRateConst", _degRateConst);
        result &= getDrugParam(drugParams, drugNodePath, "degRateConstInt", _degRateConstInt);
        result &= getDrugParam(drugParams, drugNodePath, "diffusivity", _diffusivity);
        result &= getDrugParam(drugParams, drugNodePath, "minDiffusivity", _minDiffusivity);
        result &= getDrugParam(drugParams, drugNodePath, "kAbs", _kAbs);
        result &= getDrugParam(drugParams, drugNodePath, "q", _q);
        result &= getDrugParam(drugParams, drugNodePath, "volDistCentral", _volDistCentral);
        result &= getDrugParam(drugParams, drugNodePath, "volDistPeriph", _volDistPeriph);
        result &= getDrugParam(drugParams, drugNodePath, "clearance", _clearance);
        result &= getDrugParam(drugParams, drugNodePath, "dose", _dose);
        result &= getDrugParam(drugParams, drugNodePath, "dosageInterval", _dosageInterval);
        result &= getDrugParam(drugParams, drugNodePath, "cellUptake", _cellUptake);
        result &= getDrugParam(drugParams, drugNodePath, "vascularPermeability", _vascularPermeability);

        if (_vascularPermeability*dt > 2*VOL*1000/GridSpec::MICROCMPT_AREA_2D_CENTIMETERS_SQUARED)
        {
            result = false;
        }

        result &= getDrugParam(drugParams, drugNodePath, "permCoeff", _permCoeff);
        result &= getDrugParam(drugParams, drugNodePath, "intraActivity", _intraActivity);
        result &= getDrugParam(drugParams, drugNodePath, "extraActivity", _extraActivity);
        result &= getDrugParam(drugParams, drugNodePath, "nonrepActivity", _nonrepActivity);
        result &= getDrugParam(drugParams, drugNodePath, "c50Extra", _c50Extra);
        result &= getDrugParam(drugParams, drugNodePath, "c50Intra", _c50Intra);
        result &= getDrugParam(drugParams, drugNodePath, "c50Nonrep", _c50Nonrep);
        result &= getDrugParam(drugParams, drugNodePath, "consumptionRate", _consumptionRate);
        result &= getDrugParam(drugParams, drugNodePath, "consumptionHalfsat", _consumptionHalfSat);
        result &= getDrugParam(drugParams, drugNodePath, "MIC", _mic);
        result &= getDrugParam(drugParams, drugNodePath, "hillIntra", _hillIntra);
        result &= getDrugParam(drugParams, drugNodePath, "hillExtra", _hillExtra);
        result &= getDrugParam(drugParams, drugNodePath, "hillNonrep", _hillNonrep);
        result &= getDrugParam(drugParams, drugNodePath, "caseumUnboundFraction", _caseumUnboundFraction);
        result &= getDrugParam(drugParams, drugNodePath, "caseumBindingRateConst", _caseumBindingRateConst);
        result &= getDrugParam(drugParams, drugNodePath, "epitheliumBindingKa", _epitheliumBindingKa);
        result &= getDrugParam(drugParams, drugNodePath, "epitheliumBindingRateConst", _epitheliumBindingRateConst);
        result &= getDrugParam(drugParams, drugNodePath, "cellularKout", _cellularKout);
        result &= getDrugParam(drugParams, drugNodePath, "MW", _MW);
        result &= getDrugParam(drugParams, drugNodePath, "resistRiskIndicatorFlag", _resistRiskIndicatorFlag);
        result &= getDrugParam(drugParams, drugNodePath, "aucRiskLB", _aucRiskLB);
        result &= getDrugParam(drugParams, drugNodePath, "aucRiskUP", _aucRiskUB);
        result &= getDrugParam(drugParams, drugNodePath, "cmaxRiskLB", _cmaxRiskLB);
        result &= getDrugParam(drugParams, drugNodePath, "cmaxRiskUP", _cmaxRiskUB);
        result &= getDrugParam(drugParams, drugNodePath, "numberResistMutations", _numberResistMutations);
        result &= getDrugParam(drugParams, drugNodePath, "relativeFitnessResist", _relativeFitnessResist);
        result &= getDrugParam(drugParams, drugNodePath, "numberCompensatoryMutations", _numberCompensatoryMutations);
        result &= getDrugParam(drugParams, drugNodePath, "foldIncreaseC50FromResistance", _foldIncreaseC50FromResistance);
        result &= getDrugParam(drugParams, drugNodePath, "numTimeStepsEffluxInduction", _numTimeStepsEffluxInduction);
        result &= getDrugParam(drugParams, drugNodePath, "minEffluxInduction", _minEffluxInduction);
        result &= getDrugParam(drugParams, drugNodePath, "maxEffluxInduction", _maxEffluxInduction);
        result &= getDrugParam(drugParams, drugNodePath, "concentrationChangeByEfflux", _concentrationChangeByEfflux);
        result &= getDrugParam(drugParams, drugNodePath, "numberOfRandomMonoResistantBugsToAdd", _numberOfRandomMonoResistantBugsToAdd);
        result &= getDrugParam(drugParams, drugNodePath, "volDistLung", _volDistLung);

        if(!result)
        {
            cerr << endl << "Antibiotic name " << _name
                 << " had missing or invalid Drug parameters."
                 << endl
                 << "Check that these are defined in the params.def file"
                 << " and in the parameter file."
                 << endl
                 << endl;
            throw std::runtime_error("Fatal error reading drug parameters");
        }

    }
    else
    {
        cerr << "Antibiotic name " << _name << " does not exist in the Drug section of the parameter file." << endl;
        throw std::runtime_error("Fatal error reading Drug parameters");
    }
}

void AntibioticSimulationGrid::setupDiffusion(GrDiffusion* pDiffusion, Pos dim)
{
    if (pDiffusion->getMethod() == DIFF_FFT || pDiffusion->getMethod() == DIFF_FFT_ESTIMATE)
    {
        _pMat.resize(dim.size());        
        GrDiffusionFFT* pDiffusionFFT = dynamic_cast<GrDiffusionFFT*>(pDiffusion);
        if (pDiffusionFFT)
        {
            pDiffusionFFT->setupPMATNoFlux(_pMat, _diffusivity, _PARAM(Diffusion_FFT_timestepDiffusion), _degRateConst);
        }
        else
        {
            stringstream os;
            os << "AntibioticSimulationGrid::AntibioticsimulationGrid:"
               << " Can't dynamic cast GrDiffusion* to GrDiffusionFFT*,"
               << " diffusion method: " << pDiffusion->getMethod() << endl;
            throw std::runtime_error(os.str());
        }
    }

    #if DIMENSION == 2
    // Crank-Nicolson ADI is a 2D only diffusion algorithm.
    else if(pDiffusion->getMethod() == DIFF_CNADI)
    {
        GrDiffusionCrankNicolsonADI* pDiffusionCNADI = dynamic_cast<GrDiffusionCrankNicolsonADI*>(pDiffusion);
        if (pDiffusionCNADI)
        {
            double dx = pDiffusionCNADI->_dx;
            double dt = pDiffusionCNADI->_dt;
            size_t N = pDiffusionCNADI->_N;
            size_t diagSize = pDiffusionCNADI->_diagSize;
            _cnadiChemical = new CrankNicolsonADIChemical(_diffusivity, dx, dt,
                                     N, diagSize, TribandMatrix::NEUMANN); 
        }
        else
        {
            stringstream os;
            os << "AntibioticSimulationGrid::AntibioticsimulationGrid:"
               << " Can't dynamic cast GrDiffusion* to GrDiffusionCrankNicolsonADI*,"
               << " diffusion method: " << pDiffusion->getMethod() << endl;
            throw std::runtime_error(os.str());
        }
    }
    #endif
}

template <typename T>
bool AntibioticSimulationGrid::getDrugParam(const boost::property_tree::ptree& drugParams,
                                const string& antibioticName, const string& paramName,
                                T& paramValue)
{
    boost::optional<T> optionalParamValue = drugParams.get_optional<T>(paramName);
    if(optionalParamValue)
    {
        paramValue = *optionalParamValue;
        return true;
    }
    else
    {
        cerr << "Parameter " << paramName << " not found for antibiotic " << antibioticName << endl;
        return false;
    }
}

double AntibioticSimulationGrid::getDrugParamFromLogNormalDistribution(const boost::property_tree::ptree& drugParams,
                                                                     const string& antibioticName, const string& paramName)
{
    double paramValue;
    string paramMean = paramName + "Mean";
    string paramVar = paramName + "Var";
    boost::optional<double> optionalParamMean = drugParams.get_optional<double>(paramMean);
    boost::optional<double> optionalParamVar = drugParams.get_optional<double>(paramVar);
    if(optionalParamMean && optionalParamVar)
    {
        paramValue = g_Rand.getLogNormal(*optionalParamMean,sqrt(*optionalParamVar));
        return paramValue;
    }
    else
    {
        cerr << "Mean or Variance for Parameter " << paramName << " not found for antibiotic " << antibioticName << endl;
        throw std::runtime_error("Fatal error reading Drug parameters for log normal distribution");
    }
}

void AntibioticSimulationGrid::partitionAntibiotic(GrGrid& pg, unsigned abxInd,
                               int numMolecularPerDiffusion,
                               double timestepMolecular,
                               ODESolvers::ODEMethod method)
{
    if (partitionUsingSteadyState())
    {
        partitionAntibioticSteadyState(pg, abxInd);
    }
    else
    {
        partitionAntibioticDynamic(pg, abxInd, numMolecularPerDiffusion,
                                   timestepMolecular, method);
    }
}

// Partition antibiotics on the grid based on local binding to caseum.
// Use nKillings as a proxy for the amount of caseous material in the grid
// compartment see CaseumBinding.ppt in GR-ABM-ODE/documents folder for details.
void AntibioticSimulationGrid::partitionAntibioticSteadyState(GrGrid& pg, unsigned abxInd)
{
    const Pos& dim = pg.getRange();
    GridSpec::EntireGridIterator egi = GridSpec::egiBegin(dim);
    GridSpec::EntireGridIterator egiEnd =  GridSpec::egiEnd(dim);
    for(; egi != egiEnd; ++egi)
    {
        const Pos& p = *egi;
        int nKillings = pg.nKillings(p);
        bool hasMac = pg.hasAgentType(MAC,p);

        if (!hasMac && nKillings==0)
        {
            // Get parameter for epithelium binding
            double mgFreeOld =  _currentGrid->grid(p) * VOL;                   // mg/L
            double mgEpithBndOld = _currentGrid->epithBnd_grid(p) * VOL;       // mg/L
            double mgTotalOld = mgFreeOld + mgEpithBndOld;                        // We can add these concentrations because the volumes are the same

//	                //DBG
//                        if (p == Pos(1,1))
//                        {
//                            std::cout << "----------------------------Partitioning drugs - no mac or caseation--------------------------------------" << std::endl;
//                            std::cout << "Old values---------------------" << std::endl;
//                            std::cout << "Grid position: " << p << std::endl;
//                            std::cout << "Nr of Killings: " << nKillings << std::endl;
//                            std::cout << "hasMac (0:no, 1:yes) " << hasMac << std::endl;
//                            std::cout << "epitheliumBindingKa: " << _epitheliumBindingKa << std::endl;
//                            std::cout << "mgFreeOld: " << mgFreeOld << std::endl;
//                            std::cout << "mgEpithBndOld: " << mgEpithBndOld << std::endl;
//                            std::cout << "mgTotOld: " << mgTotalOld << std::endl;
//                        }
//                        //DBG

	    double mgFreeNew = mgTotalOld / ( 1 + _epitheliumBindingKa);
            double mgEpithBndNew = mgFreeNew * _epitheliumBindingKa;
            double mgTotalNew = mgFreeNew + mgEpithBndNew;

//                        //DBG
//                        if (p == Pos(1,1))
//                        {
//                            std::cout << "New values---------------------" << std::endl;
//                            std::cout << "mgFreeNew: " << mgFreeNew << std::endl;
//                            std::cout << "mgEpithBndNew: " << mgEpithBndNew << std::endl;
//                            std::cout << "mgTotNew: " << mgTotalNew << std::endl;
//                            std::cout << "mgTotNew - mgTotOld: " << mgTotalNew - mgTotalOld << std::endl;
//                        }
                        //DBG
           
	    if (abs(mgTotalNew - mgTotalOld) > 1e-6)
            {
                cerr << setprecision(20);
                cerr << "Fatal error partitioning drugs epithelial binding---------------------------------------" << std::endl;
                cerr << "Drug name: " << _name << endl;
                cerr << "pos: " << p << endl;

                cerr << "mgFreeOld: " << mgFreeOld << endl;
                cerr << "mgEpithBndOld: " << mgEpithBndOld << endl;
                cerr << "mgFreeNew: " << mgFreeNew << endl;
                cerr << "mgEpithBndNew: " << mgEpithBndNew << endl;
                cerr << "mgFreeNew: " << setprecision(20) << mgFreeNew << endl;
                cerr << "mgEpithBndNew: " << setprecision(20) << mgEpithBndNew << endl;
                cerr << "mgFreeOld: " << setprecision(20) << mgFreeOld << endl;
                cerr << "mgEpithBndOld: " << setprecision(20) << mgEpithBndOld << endl;
                cerr << "mgTotalNew: " << setprecision(20) << mgTotalNew << endl;
                cerr << "mgTotalOld: " << setprecision(20) << mgTotalOld << endl;
                cerr << "EpitheliumBindingKa: "<< _epitheliumBindingKa << endl;
                cerr << "Difference: " << setprecision(20) << abs(mgTotalNew - mgTotalOld) << endl;
                cerr << "Difference between new and old total antibiotics is > 1e-6, epithelium binding." << endl;
                throw std::runtime_error("");
            }
            // Set new free concentration
            double freeNewConc = mgFreeNew / VOL; //mg/L
	    freeNewConc = applyCutoff(freeNewConc, _cutOffValue);
            _currentGrid->setGrid(p, freeNewConc);                         //  mg/L

            // Set new epithelium bound concentration
	    double epithBndNewConc = mgEpithBndNew / VOL;
            epithBndNewConc = applyCutoff(epithBndNewConc, _cutOffValue);
            _currentGrid->setEpithBnd_grid(p, epithBndNewConc);             //  mg/L

        }
        else
        {
            // Get parameters for caseum binding
            double uf;
            double unboundFraction;
            double dilutionFactor = 0.0;
            if (nKillings > 0)
            {
                dilutionFactor =  double(_PARAM(Core_nrKillingsCaseation)) / double(nKillings);
                if (dilutionFactor < 1) // this shouldn't happen but check just in case
                {
                    dilutionFactor = 1;
                }
                // experimentally determined unbound fraction for solid caseum
                unboundFraction = _caseumUnboundFraction;
                // grid compartment specific unbound fraction dependent on nKillings.
                uf = 1 / (1 - (1/dilutionFactor) + (1/dilutionFactor/unboundFraction));
            }
            else
            {
                uf = 1.0;     // If there's no caseum then don't partition any antibiotics there
            }

            // Get parameters for cellular uptake
            Scalar a;
            AntibioticConcentration* macAbxConc = NULL;
            double mgIntOld = 0.0;
            if (hasMac)
            {
                Mac* pMac = pg.getMac(p);
                macAbxConc = (*pMac).antibioticConcentrationList().at(abxInd);
                a = _cellUptake;
                mgIntOld = macAbxConc->concentration() * VOL_MAC;               //  mg
            }
            else
            {
                a = 0.0;                        // If there's no mac here then don't partition any antibiotics there
            }

            // Get old total antibiotic (mg) in the compartment
            // old free antibiotic + caseum bound + intracellular
            double mgFreeOld =  _currentGrid->grid(p) * VOL;                 //  mg
            double mgCasBndOld = _currentGrid->casBnd_grid(p) * VOL;         //  mg
            double mgEpithBndOld = _currentGrid->epithBnd_grid(p) * VOL;         //  mg
	    double mgTotOld = mgFreeOld + mgCasBndOld + mgIntOld + mgEpithBndOld;                   //  mg

//                        //DBG
//                        if (p == Pos(1,1))
//                        {
//                            std::cout << "----------------------------------------Partitioning drugs - mac or caseation exists---------------------------------------" << std::endl;
//                            std::cout << "Old values---------------------" << std::endl;
//                            std::cout << "Grid position: " << p << std::endl;
//                            std::cout << "Nr of Killings: " << nKillings << std::endl;
//                            std::cout << "hasMac (0:no, 1:yes) " << hasMac << std::endl;
//                            std::cout << "cellUptake: " << a << std::endl;
//                            std::cout << "unboundFraction: " << unboundFraction << std::endl;
//                            std::cout << "_PARAM(Core_nrKillingsCaseation): " << _PARAM(Core_nrKillingsCaseation) << std::endl;
//                            std::cout << "dilutionFactor: " << dilutionFactor << std::endl;
//                            std::cout << "uf: " << uf << std::endl;
//                            std::cout << "mgFreeOld: " << mgFreeOld << std::endl;
//                            std::cout << "mgCasBndOld: " << mgCasBndOld << std::endl;
//                            std::cout << "mgIntOld: " << mgIntOld << std::endl;
//                            std::cout << "mgTotOld: " << mgTotOld << std::endl;
//                        }
//                        //DBG

            // Calculate new mg amounts in each compartment
            double mgFreeNew = mgTotOld / (1 + ((1-uf)/uf) + (a*VOL_MAC/VOL));      //  mg
            double mgCasBndNew = ((1-uf)/uf) * mgFreeNew;                           //  mg
            double mgIntNew = (a*VOL_MAC/VOL) * mgFreeNew;                          //  mg
            double mgTotNew = mgFreeNew + mgCasBndNew + mgIntNew;                   //  mg

//                        //DBG
//                        if (p == Pos(1,1))
//                        {
//                            std::cout << "New values---------------------" << std::endl;
//                            std::cout << "mgFreeNew: " << mgFreeNew << std::endl;
//                            std::cout << "mgCasBndNew: " << mgCasBndNew << std::endl;
//                            std::cout << "mgIntNew: " << mgIntNew << std::endl;
//                            std::cout << "mgTotNew: " << mgTotNew << std::endl;
//                            std::cout << "mgTotNew - mgTotOld: " << mgTotNew - mgTotOld << std::endl;
//                        }
//                        //DBG
            if (abs(mgTotNew - mgTotOld) > 1e-6)
            {
                cerr << setprecision(20);
                cerr << "Fatal error partitioning drugs caseum binding---------------------------------------" << std::endl;
                cerr << "Drug name: " << _name << endl;
                cerr << "pos: " << p << endl;

                cerr << "Old values---------------------" << std::endl;
                cerr << "a: " << a << std::endl;
                cerr << "unboundFraction: " << unboundFraction << std::endl;
                cerr << "_PARAM(Core_nrKillingsCaseation): " << _PARAM(Core_nrKillingsCaseation) << std::endl;
                cerr << "dilutionFactor: " << dilutionFactor << std::endl;
                cerr << "_PARAM(Core_nrKillingsCaseation) / nKillings: " << _PARAM(Core_nrKillingsCaseation) / nKillings << std::endl;
                cerr << "nKillings: " << nKillings << std::endl;
                cerr << "uf: " << uf << std::endl;
                cerr << "mgFreeOld: " << mgFreeOld << std::endl;
                cerr << "mgCasBndOld: " << mgCasBndOld << std::endl;
                cerr << "mgIntOld: " << mgIntOld << std::endl;
                cerr << "mgTotOld: " << mgTotOld << std::endl;

                cerr << "New values---------------------" << std::endl;
                cerr << "mgFreeNew: " << mgFreeNew << std::endl;
                cerr << "mgCasBndNew: " << mgCasBndNew << std::endl;
                cerr << "mgIntNew: " << mgIntNew << std::endl;
                cerr << "mgTotNew: " << mgTotNew << std::endl;
                cerr << "mgTotNew - mgTotOld: " << mgTotNew - mgTotOld << std::endl;
                cerr << "Difference between new and old total antibiotics is > 1e-6." << endl;
                throw std::runtime_error("");
            }

            // Set new free concentration
            double freeNewConc = mgFreeNew / VOL; // mg/L
            freeNewConc = applyCutoff(freeNewConc, _cutOffValue);
            _currentGrid->setGrid(p, freeNewConc);

            // Set new caseum bound concentration
            double casBndNewConc = mgCasBndNew / VOL; // mg/L
            casBndNewConc = applyCutoff(casBndNewConc, _cutOffValue);
            _currentGrid->setCasBnd_grid(p, casBndNewConc);

            // Set new epithelium bound concentration
            _currentGrid->setEpithBnd_grid(p, 0.0);                //  mg/L

            // Set new intracellular concentration
            if (hasMac)
            {
                double intNewConc = mgIntNew / VOL_MAC; // mg/L
                intNewConc = applyCutoff(intNewConc, _cutOffValue);
                macAbxConc->setConcentration(intNewConc);
            }
        }
    }
}

// Partition antibiotics on the grid based on local binding to caseum.
// Use nKillings as a proxy for the amount of caseous material in the grid
// compartment see CaseumBinding.ppt in GR-ABM-ODE/documents folder for details.
// This version does caseum binding dynamically instead of using steady state partition.
void AntibioticSimulationGrid::partitionAntibioticDynamic(GrGrid& pg,
                               unsigned abxInd, int numMolecularPerDiffusion,
                               double timestepMolecular,
                               ODESolvers::ODEMethod method)
{
    const Pos& dim = pg.getRange();
    GridSpec::EntireGridIterator egi = GridSpec::egiBegin(dim);
    GridSpec::EntireGridIterator egiEnd =  GridSpec::egiEnd(dim);
    for(; egi != egiEnd; ++egi)
    {
        const Pos& p = *egi;
        int nKillings = pg.nKillings(p);
        bool hasMac = pg.hasAgentType(MAC,p);

        // For caseum binding
        double caseumLevel = min(double(nKillings) / double(_PARAM(Core_nrKillingsCaseation)),1.0);
        if (caseumLevel > 1.0 || caseumLevel < 0.0)
        {
            cerr << "Drug name: " << _name << endl;
            cerr << "pos: " << p << endl;
            cerr << "caseumLevel" << caseumLevel << endl;
            cerr << "nKillings" << nKillings << endl;
            cerr << "_PARAM(Core_nrKillingsCaseation)" << _PARAM(Core_nrKillingsCaseation) << endl;
            throw std::runtime_error("nKillings is below 0 or above nrKillingsCaseation - (GrSimulation::partitionAntibioticsDynamic)");
        }

        double kfc = 0.0;           // [=] (cu.s)-1 where cu are arbitrary caseum units between 0 and 1. Where cu=1 => nKillings=nrKillingsCaseation => solid caseation.
        double krc = 0.0;

        // This if statement avoids getting NaNs when caseumUnboundFraction = 1.0 (i.e. no caseum binding).
        if (_caseumUnboundFraction != 1.0)
        {
            kfc = _caseumBindingRateConst;
            double kDc = _caseumUnboundFraction / (1.0 - _caseumUnboundFraction); // Calculate kD for caseum binding. Caseum ratio of unbound to bound [=] cu, based on Vero measurements. Note that the param is the unbound FRACTION, so we have to calculate the RATIO.
            krc = kDc * kfc;     // Calculate reverse rate constant from forward rate constant and kD [=] (s)-1
        }

        // For epithelium binding
        double kfe = 0.0;           // forward rate constant for epithelium binding [=] (s)-1
        if (!hasMac && nKillings==0)
        {
            kfe = _epitheliumBindingRateConst;
        }
        double Ka = _epitheliumBindingKa;         // ratio of epithelium bound to free antibiotic (association constant) - small value to account for epithelium binding
        double kre = kfe / Ka;    // Calculate reverse rate constant from forward rate constant and Ka [=] (s)-1

        // For uptake into cells
        AntibioticConcentration* macAbxConc = NULL;
        double alpha = 0.0;                 // ratio of intracellular to extracellular (analogous to association constant Ka)
        if (hasMac)
        {
            Mac* pMac = pg.getMac(p);
            macAbxConc = (*pMac).antibioticConcentrationList().at(abxInd);
            alpha = _cellUptake;
        }
        double kout = _cellularKout;          // rate constant for drugs exiting host cells [=] (s)-1
        double kin = alpha*kout;    // calculate rate constant for entry into cells from Ka and exit rate constant [=] (s)-1


        // Solve antibiotic partitioning ODEs on molecular time step, for half of a diffusion time step
        //            cout << "Before write valarray------------------------------------" << endl;
        //            printMembers(asg, macAbxConc, p, hasMac);
        for (int MolStep = 0; MolStep < (numMolecularPerDiffusion/2); MolStep++)
        {
            updateAntibioticPartitioningODEs(macAbxConc, p, hasMac, caseumLevel, kfc, krc,
                                             kfe, kre, kin, kout, timestepMolecular, method);
        }
        //            cout << "After write valarray------------------------------------" << endl;
        //            printMembers(asg, macAbxConc, p, hasMac);
    }
}

void AntibioticSimulationGrid::updateAntibioticPartitioningODEs(AntibioticConcentration* macAbxConc, Pos p, bool hasMac,
                              double caseumLevel, double kfc, double krc, double kfe, double kre,
                              double kin, double kout, double dt, ODESolvers::ODEMethod odeSolver)
{
    valarray<double> inputVector(0.0, 4);

    // Get current values (note that we got intracellular values above)
    double freeAbx = _currentGrid->grid(p);
    double casBndAbx = _currentGrid->casBnd_grid(p);
    double epithBndAbx = _currentGrid->epithBnd_grid(p);
    double intraAbx = 0.0;
    if (hasMac)
    {
        intraAbx = macAbxConc->concentration();
    }
//    cout << "intracellular concentration before ODE is "<< intraAbx << " at pos" << p << ", hasMac" << hasMac <<endl;
//    cout << "free concentration before ODE is "<< freeAbx <<endl;
    writeValarrayFromMembersCaseumBinding(inputVector,freeAbx,casBndAbx,epithBndAbx,intraAbx);
    ODESolvers::Stepper* stepper = getStepper(odeSolver, inputVector.size());
    static GridAntibioticPartitioningFunc::Params_t params;
    params.C = &caseumLevel;
    params.kfc = &kfc;
    params.krc = &krc;
    params.kfe = &kfe;
    params.kre = &kre;
    params.kin = &kin;
    params.kout = &kout;
    GridAntibioticPartitioningFunc& fn = *getDerivFunc();//function asks for command line option bools, but they haven't been passed in this far yet!!

    stepper -> step(inputVector, fn, 0.0, dt, (void*)&params );
//    cout << "intracellular concentration after ODE is "<< inputVector[3] << " at pos" << p << ", hasMac" << hasMac <<endl;
//    cout << "kin: " << kin << endl;
//    cout << "kout: " << kout << endl;
    if (inputVector[0]<0 || inputVector[1]<0 || inputVector[2]<0 || inputVector[3]<0 ||
            isnan(inputVector[0]) || isnan(inputVector[1]) || isnan(inputVector[2]) || isnan(inputVector[3]) )
    {
        cerr << setprecision(20);
        cerr << "Fatal Error:" << endl;
        cerr << "Negative or NaN drugs -- GrSimulation::partitionAntibioticsDynamic for drug " << _name << endl;
        cerr << "Drug name: " << _name << endl;
        cerr << "pos: " << p << endl;
        cerr << "======================================" << endl;
        cerr << "inputVector[0]: " << inputVector[0] << endl;
        cerr << "inputVector[1]: " << inputVector[1] << endl;
        cerr << "inputVector[2]: " << inputVector[2] << endl;
        cerr << "inputVector[3]: " << inputVector[3] << endl;
        cerr << "===================" << endl;
        cerr << "caseumLevel: " << caseumLevel << endl;
        cerr << "kfc: " << kfc << endl;
        cerr << "krc: " << krc << endl;
        cerr << "kfe: " << kfe << endl;
        cerr << "kre: " << kre << endl;
        cerr << "kin: " << kin << endl;
        cerr << "kout: " << kout << endl;
        throw std::runtime_error("");
    }
    writeMembersFromValarrayCaseumBinding(inputVector, macAbxConc, p, hasMac);
}



void GridAntibioticPartitioningFunc::operator()(const ODESolvers::ODEState& vecread, double, ODESolvers::Derivative& vecwrite, void* params) const
{
    const double& C = *(((Params_t*)params)->C);
    const double& kfc = *(((Params_t*)params)->kfc);
    const double& krc = *(((Params_t*)params)->krc);
    const double& kfe = *(((Params_t*)params)->kfe);
    const double& kre = *(((Params_t*)params)->kre);
    const double& kin = *(((Params_t*)params)->kin);
    const double& kout = *(((Params_t*)params)->kout);

    // Free abx
    vecwrite[0] = krc*vecread[1] + kre*vecread[2] + kout*vecread[3]/VOL - (kfc*C + kfe + kin*VOL_MAC/VOL) * vecread[0];
    // caseum bound
    vecwrite[1] = kfc*vecread[0]*C - krc*vecread[1];
    // epith bound
    vecwrite[2] = kfe*vecread[0] - kre*vecread[2];
    // intracellular
    vecwrite[3] = kin*vecread[0]*VOL_MAC - kout*vecread[3];
}

void AntibioticSimulationGrid::writeValarrayFromMembersCaseumBinding(valarray<double>& inputVector, double freeAbx, double casBndAbx, double epithBndAbx, double intraAbx)
{
    // Defining the initial vector from the current grid and macrophage values

    inputVector[0] = freeAbx;
    inputVector[1] = casBndAbx;
    inputVector[2] = epithBndAbx;
    inputVector[3] = intraAbx*VOL_MAC;
}

void AntibioticSimulationGrid::writeMembersFromValarrayCaseumBinding(valarray<double>& inputVector, AntibioticConcentration* macAbxConc, Pos p, bool hasMac)
{
    _currentGrid->setGrid(p,inputVector[0]);
    _currentGrid->setCasBnd_grid(p, inputVector[1]);
    _currentGrid->setEpithBnd_grid(p, inputVector[2]);
    if (hasMac)
    {
        macAbxConc->setConcentration(inputVector[3]/VOL_MAC);
    }
}

void AntibioticSimulationGrid::printMembers(AntibioticConcentration* macAbxConc, Pos p, bool hasMac)
{
    cout << "=====================================" << endl;
    cout << "_currentGrid().grid(p)" << _currentGrid->grid(p) << endl;
    cout << "_currentGrid().casBnd_grid(p)" << _currentGrid->casBnd_grid(p) << endl;
    cout << "_currentGrid().epithBnd_grid(p)" << _currentGrid->epithBnd_grid(p) << endl;
    if (hasMac)
    {
        cout << "macAbxConc->concentration()" << macAbxConc->concentration() << endl;
    }

}



// Initialize the grid with drug concentrations specified in the parameter file
// Init section, if any.

void AntibioticSimulationGrid::initFromParameterFile(const boost::property_tree::ptree& pt)
{
    std::string initNodeName = std::string(PARAM_FILE_ROOT_NODE_NAME) + std::string(".Init");
	const boost::property_tree::ptree& initpt = pt.get_child(initNodeName);
	std::pair<boost::property_tree::ptree::const_assoc_iterator, boost::property_tree::ptree::const_assoc_iterator>
			range = initpt.equal_range(_name);

	// Iterate over each item in the Init section of the paramter file which has the node name of gridName.
	// Ex. all xml nodes in the Init section like "<TNF ... />.
	while(range.first != range.second) {
	  
		const boost::property_tree::ptree& node = (range.first++)->second;
		const boost::property_tree::ptree& p = node.get_child("<xmlattr>");
		
		// Get the value from the item.
		double value = 0.0;
		boost::optional<std::string> valueStr = p.get_optional<std::string>("value");
		if(valueStr)
		{
			stringstream ss(*valueStr);
			ss >> value;
			if(value < 0.0)
			{
				cerr << "DrugSimulationGrid::initFromParameterFile: " << _name
					 << ", invalid value of " << valueStr
					 << ", it is < 0.0" << endl;
				throw std::runtime_error("Fatal error");
			}
		}
		else
		{
			cerr << "DrugSimulationGrid::initFromParameterFile: " << _name
				 << ", no value was specified." << endl;
			throw std::runtime_error("Fatal error");
		}

		boost::optional<std::string> pos = p.get_optional<std::string>("pos");
		if(pos && *pos == std::string("all"))
		{
			// Add the specified value to each grid compartment.
            setEntireGrid(value);
		}
		else
		{
			// Add the specified value to the specified grid compartment.
			try
			{
				Pos pos = Pos(p, _currentGrid->getCenter(), _currentGrid->getSize());
                _currentGrid->setGrid(pos, value);
			}
			catch (std::exception& e)
			{
				 /* bad coordinates- get with e.what() */
				 cerr << "DrugSimulationGrid::initFromParameterFile: "
                      << _name << e.what() << endl;
				 throw std::runtime_error("Fatal error");
			}
		}
	}

}

// Resistance stats
void AntibioticSimulationGrid::updateResistanceStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_resistant._intMtb; }
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_resistant._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_resistant._nonrepExtMtb;}
    ++_resistant._totMtb;
}

void AntibioticSimulationGrid::updateMdrStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_mdr._intMtb;}
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_mdr._extMtb;}
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_mdr._nonrepExtMtb;}
    ++_mdr._totMtb;
}

void AntibioticSimulationGrid::updateCompensatoryMutationStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_compensatedResistant._intMtb;}
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_compensatedResistant._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_compensatedResistant._nonrepExtMtb; }
    ++_compensatedResistant._totMtb;
}

void AntibioticSimulationGrid::updateMdrCompensatoryMutationStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_mdrCompensatedResistant._intMtb;}
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_mdrCompensatedResistant._extMtb;}
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_mdrCompensatedResistant._nonrepExtMtb; }
    ++_mdrCompensatedResistant._totMtb;
}

void AntibioticSimulationGrid::updateMdrDoubleCompensatoryMutationStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_mdrDoubleCompensatedResistant._intMtb; }
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_mdrDoubleCompensatedResistant._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_mdrDoubleCompensatedResistant._nonrepExtMtb; }
    ++_mdrDoubleCompensatedResistant._totMtb;
}


// Cumulative resistance stats
void AntibioticSimulationGrid::updateCumulativeResistanceStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_cumResistant._intMtb;}
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_cumResistant._extMtb;}
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_cumResistant._nonrepExtMtb;}
    ++_cumResistant._totMtb;
}

void AntibioticSimulationGrid::updateCumulativeMdrStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_cumMdr._intMtb; }
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_cumMdr._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_cumMdr._nonrepExtMtb;}
    ++_cumMdr._totMtb;
}

void AntibioticSimulationGrid::updateCumulativeCompensatoryMutationStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_cumCompensated._intMtb; }
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_cumCompensated._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_cumCompensated._nonrepExtMtb; }
    ++_cumCompensated._totMtb;
}

void AntibioticSimulationGrid::updateCumulativeMdrCompensatoryMutationStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_cumMdrCompensatedResistant._intMtb; }
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_cumMdrCompensatedResistant._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_cumMdrCompensatedResistant._nonrepExtMtb; }
    ++_cumMdrCompensatedResistant._totMtb;
}

void AntibioticSimulationGrid::updateCumulativeMdrDoubleCompensatoryMutationStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_cumMdrDoubleCompensated._intMtb;}
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_cumMdrDoubleCompensated._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_cumMdrDoubleCompensated._nonrepExtMtb; }
    ++_cumMdrDoubleCompensated._totMtb;
}

void AntibioticSimulationGrid::updateEffluxInducedStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_effluxInduced._intMtb; }
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_effluxInduced._extMtb;}
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_effluxInduced._nonrepExtMtb; }
    ++_effluxInduced._totMtb;
}

void AntibioticSimulationGrid::updateDoubleEffluxInducedStats(MtbGroupLocation location)
{
    if (location == INTRA_CELLULAR)  {
        ++_doubleEffluxInduced._intMtb;}
    if (location == REPLICATING_EXTRA_CELLULAR || location == NON_REPLICATING_EXTRA_CELLULAR)  {
        ++_doubleEffluxInduced._extMtb; }
    if (location == NON_REPLICATING_EXTRA_CELLULAR) {
        ++_doubleEffluxInduced._nonrepExtMtb; }
    ++_doubleEffluxInduced._totMtb;
}
